import aiohttp
import random
import abc
from typing import Dict, Collection, Optional
from .typedefs import Proxy, Session


class ProxyPool:
    def __init__(self, proxies: Collection[Proxy]):
        self._proxies = set(proxies)

    def rand(self) -> Optional[Proxy]:
        if len(self._proxies) == 0:
            return None
        return random.sample(self._proxies, 1)[0]

    def pop(self, proxy: str) -> None:
        if proxy in self._proxies:
            self._proxies.remove(proxy)


class SessionPool(abc.ABC):
    @abc.abstractmethod
    def rand(self) -> Session:
        """ """

    @abc.abstractmethod
    def pop(self, key: Optional[Proxy]) -> None:
        """ """


class ProxySessionPool(SessionPool):
    def __init__(self, proxy_pool: ProxyPool, size: int, session_kwargs: dict = None, cookies: dict = None):
        self._size = size
        self._proxy_pool = proxy_pool
        self._session_kwargs = session_kwargs or {}
        self._cookies = cookies or {}
        self._session_pool: Dict[Proxy, aiohttp.ClientSession] = {}
        for _ in range(self._size):
            self._add_session()

    def rand(self) -> Session:
        if len(self._session_pool) == 0:
            raise IndexError('No sessions left')
        return random.choice(list(self._session_pool.items()))

    def pop(self, key: Optional[Proxy]) -> None:
        if key in self._session_pool:
            self._session_pool.pop(key)
            self._add_session()

    def _add_session(self) -> None:
        proxy = self._proxy_pool.rand()
        if proxy is not None:
            session_kwargs = self._session_kwargs
            if proxy in self._cookies:
                session_kwargs["cookies"] = self._cookies[proxy]
            self._session_pool[proxy] = aiohttp.ClientSession(**session_kwargs)
            self._proxy_pool.pop(proxy)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for session in self._session_pool.values():
            await session.close()


class SingleSessionPool(SessionPool):
    def __init__(self, session_kwargs: dict = None):
        if session_kwargs is None:
            session_kwargs = {}
        self.session = (None, aiohttp.ClientSession(**session_kwargs))

    def rand(self) -> Session:
        return self.session

    def pop(self, key: Optional[Proxy]) -> None:
        pass



from .client import (
    WebTextClient,
    WebByteClient,
    Client,
    CacheClient,
    CacheOnlyClient,
    CacheSkipClient,
    RetryClient,
    CrawlerClient,
    WebClient,
    ImageClient
)

from .cache import (
    Cache,
    FileCache
)

from .html import (
    select_one,
    select_text_one,
    select_all
)

from .session import (
    ProxyPool,
    SessionPool,
    SingleSessionPool,
    ProxySessionPool
)

from .worker import (
    Dispatcher,
    Worker,
    Master,
    CrawlerWorker,
    SimpleWorker
)



import os
import sys
from shutil import rmtree

from setuptools import setup, Command

# Package meta-data.
NAME = "aioscrapy"
DESCRIPTION = "Python asynchronous library for web scraping"
URL = "https://github.com/eugen1j/aioscrapy"
EMAIL = "eugenij.bondar@gmail.com"
AUTHOR = "eugen1j"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "0.1.10"

here = os.path.abspath(os.path.dirname(__file__))
with open(f"{here}/README.md") as f:
    long_description = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel distribution...")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    packages=[NAME],
    install_requires=[
        "pyfakefs",
        "pytest",
        "beautifulsoup4",
        "aiohttp",
        "pytest-asyncio",
        "lxml",
        "pytest-aiohttp",
    ],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_dir={NAME: NAME},
    include_package_data=True,
    url=URL,
    author_email=EMAIL,
    license="MIT",
    keywords="""
                python3
                aioscrapy
                crawler
                scraper
                python-library
               """,
    python_requires=REQUIRES_PYTHON,
    zip_safe=False,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
    ],
    # Build and upload package: python3 setup.py upload
    cmdclass={"upload": UploadCommand},
)



import hashlib
import os
import abc
import pickle
from typing import Generic, Optional
from .typedefs import VT, KT


class Cache(abc.ABC, Generic[KT, VT]):
    @abc.abstractmethod
    def get(self, key: KT) -> Optional[VT]:
        """ """

    @abc.abstractmethod
    def set(self, key: KT, val: VT) -> None:
        """ """


class FileCache(Cache[str, VT]):
    def __init__(self, folder: str):
        self._folder = folder

    def get(self, key: str) -> Optional[VT]:
        path = self._full_path(key)
        try:
            with open(path, 'rb') as file:
                return pickle.load(file)
        except FileNotFoundError:
            return None

    def set(self, key: str, val: VT) -> None:
        path = self._full_path(key)
        directory = os.path.dirname(path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(path, 'wb') as file:
            pickle.dump(val, file, pickle.HIGHEST_PROTOCOL)

    def _full_path(self, key: str) -> str:
        md5 = hashlib.md5(key.encode()).hexdigest()
        return os.path.join(self._folder, md5[:2], md5)


class FakeCache(Cache[KT, VT]):
    def __init__(self):
        self._cache = {}

    def get(self, key: KT) -> Optional[VT]:
        return self._cache.get(key, None)

    def set(self, key: KT, val: VT) -> None:
        self._cache[key] = val
